@@ -883,7 +883,7 @@ gst_ffmpegvidenc_set_property (GObject * object,
 
   if (ffmpegenc->context) {
     GST_WARNING_OBJECT (ffmpegenc,
-        "Can't change properties once decoder is setup !");
+        "Can't change properties once encoder is setup !");
     return;
   }
 
@@ -1024,6 +1024,11 @@ gst_ffmpegvidenc_register (GstPlugin * plugin)
     if (in_plugin->type != AVMEDIA_TYPE_VIDEO)
       continue;
 
+    /* Skip formats we don't handle */
+    if (!gst_ffmpeg_codecid_is_known (in_plugin->id)) {
+      continue;
+    }
+
     /* no quasi codecs, please */
     if (in_plugin->id == AV_CODEC_ID_RAWVIDEO ||
         in_plugin->id == AV_CODEC_ID_V210 ||
@@ -1087,6 +1092,12 @@ gst_ffmpegvidenc_register (GstPlugin * plugin)
       continue;
     }
 
+    int rank = GST_RANK_SECONDARY;
+
+    if (!strcmp (in_plugin->name, "hap")) {
+      rank = GST_RANK_NONE;
+    }
+
     /* construct the type */
     type_name = g_strdup_printf ("avenc_%s", in_plugin->name);
 
@@ -1110,7 +1121,7 @@ gst_ffmpegvidenc_register (GstPlugin * plugin)
       }
     }
 
-    if (!gst_element_register (plugin, type_name, GST_RANK_SECONDARY, type)) {
+    if (!gst_element_register (plugin, type_name, rank, type)) {
       g_free (type_name);
       return FALSE;
     }
